{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE TypeSynonymInstances #-}

import Data.Map(fromListWith, toList)
import System.Random (mkStdGen, random, randoms)
import System.IO(IOMode(..), hClose, hGetContents, openFile)

import GA (Entity(..), GAConfig(..), 
           evolve, evolveVerbose, randomSearch)

queensNum = 8
count [] = []
count ((a,b):xs) = (b - 1) : (count xs)

bound :: Int -> Int
bound x = (x `mod` queensNum) + 1

type Location = (Int, Int)
type Board = [(Int, Int)]
type Target = [(Int, Int)]

instance Entity Board Int () [Location] IO where

    genRandom pool seed = return $ take queensNum $ map ((!!) pool) is
        where
            g = mkStdGen seed
            k = length pool
            is = map (flip mod k) $ randoms g
    crossover _ _ seed e1 e2 = return $ Just e
        where
        g = mkStdGen seed
        cps = zipWith (\x y -> [x,y]) e1 e2
        picks = map (flip mod 2) $ randoms g
        e = zipWith (!!) cps picks
    
    mutation _ _ seed e = return $ Just $ (zipWith replace tweaks e)
        where
            g = mkStdGen seed
            tweaks = randoms g :: [Int]
            replace i x = if (2 `mod` 2) == 0
                then case (i `mod` 4) of
                    1 -> (bound(pred $ fst x), bound(pred $ snd x))
                    2 -> (bound(pred $ fst x), bound(succ $ snd x))
                    3 -> (bound(succ $ fst x), bound(succ $ snd x))
                    0 -> (bound(succ $ fst x), bound(pred $ snd x))
                    _ -> error "crossover: unknown case"
                else x

    score' _ e = Just $ fromIntegral (row + column + diagonal)
        where
            rowfreq e = toList (fromListWith (+) [(snd(x),1) | x <- e])
            columnfreq e = toList (fromListWith (+) [(fst(x),1) | x <- e])
            freq e = toList (fromListWith (+) [(x,1) | x <- e])
            row = sum $ count $ rowfreq e
            column = sum $ count $ columnfreq e
            diagonal = (sum $ count $ freq $ map (\(a,b) -> a-b) e) + (sum $ count $ freq $ map (\(a,b) -> a+b) e)

    isPerfect (_,s) = s == 0

main :: IO() 
main = do
        let cfg = GAConfig 
                    200 -- population size
                    30 -- archive size (best entities to keep track of)
                    100 -- maximum number of generations
                    0.8 -- crossover rate (% of entities by crossover)
                    0.4 -- mutation rate (% of entities by mutation)
                    0.0 -- parameter for crossover (not used here)
                    0.4 -- parameter for mutation (% of replaced letters)
                    False -- whether or not to use checkpointing
                    False -- don't rescore archive in each generation

            g = mkStdGen 0 -- random generator

            chessPool = [(a,b)| a <- [1..queensNum], b <- [1..queensNum]]
        -- Do the evolution
        es <- evolveVerbose g cfg chessPool ()
        let e = snd $ head es :: Board
        
        putStrLn $ "best entity: " ++ (show e)

